// Copyright (c) 2020 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "source/fuzz/pass_management/repeated_pass_manager_simple.h"

namespace spvtools {
namespace fuzz {

RepeatedPassManagerSimple::RepeatedPassManagerSimple(
    FuzzerContext* fuzzer_context, RepeatedPassInstances* pass_instances)
    : RepeatedPassManager(fuzzer_context, pass_instances) {}

RepeatedPassManagerSimple::~RepeatedPassManagerSimple() = default;

FuzzerPass* RepeatedPassManagerSimple::ChoosePass(
    const protobufs::TransformationSequence& /*unused*/) {
  auto& passes = GetPassInstances()->GetPasses();
  return passes[GetFuzzerContext()->RandomIndex(passes)].get();
}

}  // namespace fuzz
}  // namespace spvtools
